#include "MsdfAtlasLayerCombo.h"
#include "AtlasLayerCombo.h"
#include "MsdfLayer.h"
#include "Utils.h"
#include "ofFileUtils.h"
#include <GL/glext.h>
#include <GLES3/gl3.h>

namespace ofxVariableLab {

void MsdfAtlasLayerCombo::setup(const ComboIdentifier & layerIdentifier,
                                const AtlasLayerComboSettings & settings){
    this->identifier = layerIdentifier;
    this->settings = static_cast <const MsdfAtlasLayerComboSettings &>(settings);
    ofxMsdfgen::AtlasSettings atlasSettings;
    //settings.characters = "ABCDEFGHIJKL";
    atlasSettings.scale = this->settings.scale;
    atlasSettings.minimumScale = this->settings.minimumScale;
    atlasSettings.characters = "";
    atlasSettings.maxInterpolationStepSize = this->settings.maxInterpolationStepSize;
    scaleFactor = 64.0 / this->settings.scale;

    atlas = make_shared <ofxMsdfgen::Atlas>();

    msdfShader = make_shared <ofShader>();
    #ifdef TARGET_EMSCRIPTEN
        if(ofFile("shaders/ES3/msdf/shader.frag").exists()){
            msdfShader->load("shaders/ES3/msdf/shader");
        }else{
            msdfShader->load("ofxMsdfgen/shaders/mix/ES3/shader");
        }
    #else
        msdfShader->load("ofxMsdfgen/shaders/mix/GL3/shader");
    #endif

    atlas->setup(this->identifier.fontPath, atlasSettings);
}

void MsdfAtlasLayerCombo::update(){
    if(isDirty){
        bool useCache = true;
        bool saveToCache = true;
        atlas->generate(useCache, saveToCache);
        isDirty = false;
    }
    if(!uboIsSetup){
        setupUbo();
    }
}

void MsdfAtlasLayerCombo::setupUbo(){
    msdfShader->begin();
    GLint prog = 0;
    glGetIntegerv(GL_CURRENT_PROGRAM, &prog);
    msdfShader->end();

    glGenBuffers(1, &uboName);
    uboIndex = glGetUniformBlockIndex(prog, "ubo");
    glBindBufferBase(GL_UNIFORM_BUFFER, 0, uboName);
    MsdfLayer::UboParameters up;
    glBufferData(GL_UNIFORM_BUFFER, sizeof(glm::vec4) * 8, NULL, GL_DYNAMIC_DRAW);
    glUniformBlockBinding(prog, uboIndex, 0);
    for(const auto & layer : layers){
        layer->uboName = uboName;
        layer->uboIndex = uboIndex;
    }
    uboIsSetup = true;
}

void MsdfAtlasLayerCombo::careForChild(shared_ptr <Layer> layer){
    shared_ptr <MsdfLayer> msdfLayer = dynamic_pointer_cast <MsdfLayer>(layer);
    msdfLayer->setMomsComboIdentifier(identifier);
    msdfLayer->isWithNewMom();
    msdfLayer->setDirtyDirty(true);
    msdfLayer->setAtlas(this->atlas);
    msdfLayer->setShader(msdfShader);
    msdfLayer->uboName = uboName;
    msdfLayer->uboIndex = uboIndex;
    msdfLayer->scaleFactor = scaleFactor;
    auto & as = atlas->settings;
    for(const char c : layer->getProps().text){
        if(as.characters.find(c) == std::string::npos){
            // not found
            as.characters += c;
            isDirty = true;
        }
    }
    layers.push_back(std::move(msdfLayer));
}
void MsdfAtlasLayerCombo::abandonChild(shared_ptr <Layer> layer){
    shared_ptr <MsdfLayer> msdfLayer = dynamic_pointer_cast <MsdfLayer>(layer);
    layers.erase(
        std::remove_if(
            layers.begin(),
            layers.end(),
            [msdfLayer](shared_ptr <MsdfLayer> const & l)
        {
            return l == msdfLayer;
        }), layers.end());
}

const ComboIdentifier & MsdfAtlasLayerCombo::getIdentifier() const {
    return identifier;
}

void MsdfAtlasLayerCombo::setVFlip(VFlipState vFlipState){
    vFlip = vFlipState;
    for(const auto & layer : layers){
        layer->setVFlip(vFlip);
    }
}

bool MsdfAtlasLayerCombo::hasChildren(){
    return layers.size() > 0;
}

const ofImage & MsdfAtlasLayerCombo::getAtlasImage(){
    return atlas->getAtlasImage();
}

string MsdfAtlasLayerCombo::getAtlasImagePath(){
    return atlas->getAtlasPath();
}
}
